'''
TRXASprefitpack:
package for TRXAS pre- and fitting process which aims for the first order dynamics
TRXAS stands for Time Resolved X-ray Absorption Spectroscopy

:copyright: 2021-2022 by pistack (Junho Lee)
:license: LGPL3.
'''

from .mathfun import *
from .driver import *
from .res import *

__version__ = '0.7.2'
